classdef AprilTagFinder < handle
% A matlab wrapper class for easy access video stream and apriltag.
% To construct an AprilTagFinder object and start internal video
% capturer, use any of the following forms:
% h=AprilTagFinder; %default Tag36h11, video 0, and show window
% h=AprilTagFinder(tagfamilyName);
% h=AprilTagFinder(tagfamilyName,videoDevice);
% h=AprilTagFinder(tagfamilyName,videoDevice, showColorWindow);
%
% author: Chen Feng <simbaforrest at gmail dot com>

  %% private members
  properties (SetAccess = private)
    tagfamily
    finder
  end
  
  %% public methods
  methods
    function h=AprilTagFinder(varargin)
      %construct an AprilTagFinder object and start internal video
      %capturer using any of the following forms:
      % h=AprilTagFinder; %default
      % h=AprilTagFinder(tagfamilyName);
      % h=AprilTagFinder(tagfamilyName,videoDevice);
      % h=AprilTagFinder(tagfamilyName,videoDevice, showColorWindow);
      Apriltag4Matlab.install; %ensure jar installed
      
      %default input args
      tagfamilyName='Tag36h11';
      videoDevice=0;
      showColorWindow=true;
      switch nargin
        case 1
          tagfamilyName=varargin{1};
        case 2
          [tagfamilyName, videoDevice]=deal(varargin{:});
        case 3
          [tagfamilyName, videoDevice, showColorWindow]=deal(varargin{:});
      end
      fprintf('AprilTagFinder to be started with:\n');
      fprintf('\tTagfamily: %s\n',tagfamilyName);
      if ischar(videoDevice)
        fprintf('\tVideoDevice: %s\n', videoDevice);
      else
        fprintf('\tVideoDevice: %d\n', videoDevice);
      end
      fprintf('\tShowColorWindow: %d\n',showColorWindow);
      
      try
        h.tagfamily=javaObject(['april.tag.',tagfamilyName]);
        h.finder=matlab.apriltag.AprilTagFinder(h.tagfamily);
        h.finder.start(videoDevice, showColorWindow);
      catch err
        error(err.message);
      end
    end
    
    function [ds,img]=find(h)
      % find apriltags, return:
      %  ds: an array of detections, empty if none
      %  img: the current image frame in grey mode (an uint8 matrix)
      try
        result=h.finder.find();
      catch err
        error(err.message);
      end
      
      if nargout>0
        raw = result.detections;
        nds=raw.size();
        ds=cell(nds,1);
        for i=0:(nds-1)
          rawd=raw.get(i);
          d.p=rawd.p';
          d.cxy=rawd.cxy;
          d.H=rawd.homography;
          d.id=rawd.id;
          d.hammingDistance=rawd.hammingDistance;
          ds{i+1}=d;
        end
        ds=cell2mat(ds);
      end
      
      if nargout>1
        fimg = result.frame;
        gimg=reshape(fimg.d,fimg.width,fimg.height)';
        img=uint8(255*gimg);
      end
    end
    
    function delete(h)
      %stop internal video capturer
      %usually, no need to call this directly, will be automatically called
      %when h is cleared
      assert(isa(h,'AprilTagFinder'));
      try
        h.finder.close();
      catch err
        error(err.message);
      end
    end
    
    function key=waitkey(h,ms)
      if isempty(h.finder.canvas)
        disp('[AprilTagFinder.waitkey warn]: you did not open java canvas!');
        key='';
        return;
      end
      ke=h.finder.canvas.waitKey(ms);
      if ~isempty(ke)
        key=ke.getKeyCode();
      else
        key=-eps;
      end
    end
  end
  
  %% static methods
  methods(Static)    
    function plot(detections)
      holded=ishold;
      hold on;
      nds=length(detections);
      for i=1:nds
        d=detections(i);
        d.p=d.p+1; %note matlab draw image using 1-base index
        d.cxy=d.cxy+1;
        drawTag(d);
      end
      if ~holded; hold off; end;
    end
  end
end

%% private helper functions
function drawLine(st, ed, cl)
plot([st(1), ed(1)], [st(2), ed(2)], 'Color', cl, 'LineWidth', 2);
end

function drawLines(vertex, indices, colors)
for i=1:length(colors)
  drawLine(vertex(:,indices(i,1)), vertex(:,indices(i,2)), colors(i,:));
end
end

function drawTag(d)
drawLines(d.p, [1,2;2,3;3,4;4,1;1,3;2,4], [0,0,0;1,0,0;0,1,0;0,0,1;1,1,0;1,1,0]);
text(d.cxy(1), d.cxy(2), num2str(d.id));
end